@@ -5,6 +5,8 @@
 """
 
 
+from __future__ import unicode_literals
+
 import logging
 from inspect import ismethod
 
@@ -12,6 +14,7 @@ from django.core.urlresolvers import (reverse, resolve, NoReverseMatch,
                                       Resolver404)
 from django.utils.html import escape
 from django.utils.safestring import mark_safe
+from django.utils.encoding import smart_text
 from django.utils.translation import ugettext as _
 from django.db.models import Model
 from django import template
@@ -25,7 +28,18 @@ register = template.Library()
 CONTEXT_KEY = 'DJANGO_BREADCRUMB_LINKS'
 
 
-def breadcrumb(context, label, viewname, *args):
+def append_breadcrumb(context, label, viewname, args, kwargs):
+    if 'request' in context:
+        context['request'].META[CONTEXT_KEY] = context['request'].META.get(
+            CONTEXT_KEY, []) + [(label, viewname, args, kwargs)]
+    else:
+        logger.error("request object not found in context! Check if "
+                     "'django.core.context_processors.request' is in "
+                     "TEMPLATE_CONTEXT_PROCESSORS")
+
+
+@register.simple_tag(takes_context=True)
+def breadcrumb(context, label, viewname, *args, **kwargs):
     """
     Add link to list of breadcrumbs, usage:
 
@@ -40,35 +54,43 @@ def breadcrumb(context, label, viewname, *args):
                      instance with implemented get_absolute_url().
     :param args: Any arguments to view function.
     """
-    if 'request' in context:
-        context['request'].META[CONTEXT_KEY] = context['request'].META.get(
-            CONTEXT_KEY, []) + [(escape(label), viewname, args)]
-    else:
-        logger.error("request object not found in context! Check if "
-                     "'django.core.context_processors.request' is in "
-                     "TEMPLATE_CONTEXT_PROCESSORS")
+    append_breadcrumb(context, _(escape(label)), viewname, args, kwargs)
     return ''
 
 
-def breadcrumb_safe(context, label, viewname, *args):
+@register.simple_tag(takes_context=True)
+def breadcrumb_safe(context, label, viewname, *args, **kwargs):
     """
     Same as breadcrumb but label is not escaped.
     """
-    if 'request' in context:
-        context['request'].META[CONTEXT_KEY] = context['request'].META.get(
-            CONTEXT_KEY, []) + [(label, viewname, args)]
-    else:
-        logger.error("request object not found in context! Check if "
-                     "'django.core.context_processors.request' is in "
-                     "TEMPLATE_CONTEXT_PROCESSORS")
+    append_breadcrumb(context, _(label), viewname, args, kwargs)
     return ''
 
 
+@register.simple_tag(takes_context=True)
+def breadcrumb_raw(context, label, viewname, *args, **kwargs):
+    """
+    Same as breadcrumb but label is not translated.
+    """
+    append_breadcrumb(context, escape(label), viewname, args, kwargs)
+    return ''
+
+
+@register.simple_tag(takes_context=True)
+def breadcrumb_raw_safe(context, label, viewname, *args, **kwargs):
+    """
+    Same as breadcrumb but label is not escaped and translated.
+    """
+    append_breadcrumb(context, label, viewname, args, kwargs)
+    return ''
+
+
+@register.simple_tag(takes_context=True)
 def render_breadcrumbs(context, *args):
     """
-    Render breadcrumbs html using twitter bootstrap css classes.
+    Render breadcrumbs html using bootstrap css classes.
     """
-    if not 'request' in context:
+    if 'request' not in context:
         logger.error("request object not found in context! Check if "
                      "'django.core.context_processors.request' is in "
                      "TEMPLATE_CONTEXT_PROCESSORS")
@@ -80,12 +102,12 @@ def render_breadcrumbs(context, *args):
         template_path = 'django_bootstrap_breadcrumbs/bootstrap2.html'
 
     links = []
-    for (label, viewname, view_args) in context['request'].META.get(
-            CONTEXT_KEY, []):
+    for (label, viewname, view_args, view_kwargs) in context[
+            'request'].META.get(CONTEXT_KEY, []):
         if isinstance(viewname, Model) and hasattr(
                 viewname, 'get_absolute_url') and ismethod(
                 viewname.get_absolute_url):
-            url = viewname.get_absolute_url()
+            url = viewname.get_absolute_url(*view_args, **view_kwargs)
         else:
             try:
                 try:
@@ -98,10 +120,10 @@ def render_breadcrumbs(context, *args):
                     except Resolver404:
                         current_app = None
                 url = reverse(viewname=viewname, args=view_args,
-                              current_app=current_app)
+                              kwargs=view_kwargs, current_app=current_app)
             except NoReverseMatch:
                 url = viewname
-        links.append((url, _(unicode(label)) if label else label))
+        links.append((url, smart_text(label) if label else label))
 
     if not links:
         return ''
@@ -116,10 +138,17 @@ class BreadcrumbNode(template.Node):
     def __init__(self, nodelist, viewname, args):
         self.nodelist = nodelist
         self.viewname = viewname
-        self.args = args
+        self.args = list(args)
+        self.kwargs = {}
+        for arg in args:
+            if '=' in arg:
+                name = arg.split('=')[0]
+                val = '='.join(arg.split('=')[1:])
+                self.kwargs[name] = val
+                self.args.remove(arg)
 
     def render(self, context):
-        if not 'request' in context:
+        if 'request' not in context:
             logger.error("request object not found in context! Check if "
                          "'django.core.context_processors.request' is in "
                          "TEMPLATE_CONTEXT_PROCESSORS")
@@ -130,8 +159,8 @@ class BreadcrumbNode(template.Node):
         except template.VariableDoesNotExist:
             viewname = self.viewname
         args = self.parse_args(context)
-        context['request'].META[CONTEXT_KEY] = context['request'].META.get(
-            CONTEXT_KEY, []) + [(label, viewname, args)]
+        kwargs = self.parse_kwargs(context)
+        append_breadcrumb(context, label, viewname, args, kwargs)
         return ''
 
     def parse_args(self, context):
@@ -144,7 +173,18 @@ class BreadcrumbNode(template.Node):
             args.append(value)
         return args
 
+    def parse_kwargs(self, context):
+        kwargs = {}
+        for name, val in self.kwargs.items():
+            try:
+                value = template.Variable(val).resolve(context)
+            except template.VariableDoesNotExist:
+                value = val
+            kwargs[name] = value
+        return kwargs
 
+
+@register.tag
 def breadcrumb_for(parser, token):
     bits = list(token.split_contents())
     end_tag = 'end' + bits[0]
@@ -153,11 +193,12 @@ def breadcrumb_for(parser, token):
     return BreadcrumbNode(nodelist, bits[1], bits[2:])
 
 
+@register.simple_tag(takes_context=True)
 def clear_breadcrumbs(context, *args):
     """
     Removes all currently added breadcrumbs.
     """
-    if not 'request' in context:
+    if 'request' not in context:
         logger.error("request object not found in context! Check if "
                      "'django.core.context_processors.request' is in "
                      "TEMPLATE_CONTEXT_PROCESSORS")
@@ -167,10 +208,3 @@ def clear_breadcrumbs(context, *args):
         del context['request'].META[CONTEXT_KEY]
 
     return ''
-
-
-register.simple_tag(takes_context=True)(breadcrumb)
-register.simple_tag(takes_context=True)(breadcrumb_safe)
-register.simple_tag(takes_context=True)(render_breadcrumbs)
-register.simple_tag(takes_context=True)(clear_breadcrumbs)
-register.tag('breadcrumb_for', breadcrumb_for)
